import os
import shutil
import subprocess
import re
import logging
from pathlib import Path

import click
from rich.console import Console
from rich.logging import RichHandler
from rich.progress import track

from compress_mail._version import __version__
from compress_mail.lockmaildir import maildir_lock


logger = logging.getLogger(__name__)
console = Console(highlight=False)


def sizeof_fmt(
    num: float,
    suffix: str = 'B',
    units: list[str] = None,
    power: int = None,
    sep: str = '',
    precision: int = 2,
    sign: bool = False,
) -> str:
    sign = '+' if sign and num > 0 else ''
    fmt = '{0:{1}.{2}f}{3}{4}{5}'
    prec = 0
    for unit in units[:-1]:
        if abs(round(num, precision)) < power:
            break
        num /= float(power)
        prec = precision
    else:
        unit = units[-1]
    return fmt.format(num, sign, prec, sep, unit, suffix)


def sizeof_fmt_iec(num: float, suffix: str = 'B', sep: str = '', precision: int = 2, sign: bool = False) -> str:
    return sizeof_fmt(
        num,
        suffix=suffix,
        sep=sep,
        precision=precision,
        sign=sign,
        units=['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'],
        power=1024,
    )


def sizeof_fmt_decimal(num: float, suffix: str = 'B', sep: str = '', precision: int = 2, sign: bool = False) -> str:
    return sizeof_fmt(
        num,
        suffix=suffix,
        sep=sep,
        precision=precision,
        sign=sign,
        units=['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'],
        power=1000,
    )


class MailCompressor:
    def __init__(
        self, maildir: str, tmp_dir: str, control_dir: str, timeout: int, compression_method: str, use_lock: bool
    ) -> object:
        self.maildir: Path = Path(maildir)
        self.tmp_dir: Path = Path(tmp_dir)
        self.control_dir: Path = Path(control_dir)
        self.timeout = timeout
        self.compression_method = compression_method
        self.use_lock = use_lock
        self.pid = None

    def check_binaries(self) -> None:
        required_binaries = [self.compression_method]
        for binary in required_binaries:
            if not shutil.which(binary):
                raise FileNotFoundError(f'Required binary not found: {binary}')
        logger.info('All required binaries are present')

    def find_mails_to_compress(self) -> list[Path]:
        mails_to_compress: list[Path] = []
        for root, dirs, files in os.walk(self.maildir):
            regex_non_match_count = 0
            z_missing_count = 0
            already_compressed = 0
            for filename in track(files, description=f'{"Searching for mails":<25}'):
                if re.search(r'S=\d+', filename):
                    if 'Z' not in filename.split(':')[1]:
                        mails_to_compress.append(Path(root) / filename)
                    else:
                        already_compressed += 1
                else:
                    regex_non_match_count += 1
            logger.info(
                f'Matched mails: {len(mails_to_compress)}, already compressed: {already_compressed}, missing size field: {regex_non_match_count}'
            )

        if mails_to_compress:
            logger.info(f'Found {len(mails_to_compress)} mails to compress')
        else:
            logger.info('No mails found to compress')
        return mails_to_compress

    def compress_mails(self, mails: list[Path]) -> list[Path]:
        compressed_files: list[Path] = []
        for mail in track(mails, description=f'{"Compressing mails":<25}'):
            compressed_file = self.tmp_dir / mail.name
            logger.debug(f'Compressing mail: {mail}')
            logger.debug(f'Writing to file: {compressed_file}')
            shutil.copy2(mail, compressed_file)
            if self.compression_method == 'gzip':
                subprocess.run(['gzip', '-6', compressed_file])
                compressed_files.append(self.with_ending(compressed_file, '.gz'))
            elif self.compression_method == 'zstd':
                subprocess.run(['zstd', '-3', '--rm', '-q', '-T0', compressed_file])
                compressed_files.append(self.with_ending(compressed_file, '.zst'))
        logger.info(f'Compressed {len(compressed_files)} mails using {self.compression_method}')
        return compressed_files

    def update_mtime(self, original_files: list[Path], compressed_files: list[Path]) -> None:
        zipped = list(zip(original_files, compressed_files))
        for original, compressed in track(zipped, description=f'{"Updating mtime":<25}'):
            logger.debug(f'Original file: {original}')
            logger.debug(f'Compressed file: {compressed}')
            original_mtime = original.stat().st_mtime
            os.utime(compressed, (original_mtime, original_mtime))
        logger.info('Updated mtimes for compressed files')

    def verify_and_replace_mails(self, original_files: list[Path], compressed_files: list[Path]) -> None:
        zipped = list(zip(original_files, compressed_files))
        for original, compressed in track(zipped, description=f'{"Verifying/replacing mails":<25}'):
            logger.debug(f'Original file: {original}')
            logger.debug(f'Compressed file: {compressed}')
            if original.exists():
                final = compressed.replace(self.with_ending(original, 'Z'))
                logger.debug(f'Final file: {final}')
                original.unlink()
            else:
                compressed.unlink()
        logger.info('Verified and replaced original mails with compressed ones')

    def get_directory_size(self, directory: Path) -> int:
        total_size = 0
        for dirpath, dirnames, filenames in os.walk(directory):
            for filename in filenames:
                file_path = Path(dirpath) / filename
                total_size += file_path.stat().st_size
        return total_size

    def with_ending(self, path: Path, ending: str) -> Path:
        return path.with_name(path.name + ending)

    def run(self):
        try:
            logger.info('Starting maildir compression process')
            logger.info(f'maildir: {self.maildir}')
            logger.info(f'tmp_dir: {self.tmp_dir}')
            logger.info(f'control_dir: {self.control_dir}')

            self.check_binaries()
            initial_size = self.get_directory_size(self.maildir)
            logger.info(f'Initial maildir size: {sizeof_fmt_decimal(initial_size)}')

            mails_to_compress = self.find_mails_to_compress()
            if not mails_to_compress:
                return
            logger.debug(f'Found {mails_to_compress} mails to compress')

            compressed_files = self.compress_mails(mails_to_compress)
            self.update_mtime(mails_to_compress, compressed_files)

            if self.use_lock:
                with maildir_lock(self.control_dir, self.timeout):
                    self.verify_and_replace_mails(mails_to_compress, compressed_files)
            else:
                self.verify_and_replace_mails(mails_to_compress, compressed_files)

            final_size = self.get_directory_size(self.maildir)
            logger.info(f'Final maildir size: {sizeof_fmt_decimal(final_size)}')

            savings = ((initial_size - final_size) / initial_size) * 100
            logger.info(f'Disk savings: {savings:.2f}%')

            logger.info('Completed maildir compression process')

        except FileNotFoundError as e:
            logger.error(f'Error occurred: {e}')
            console.print_exception(show_locals=True, width=250, word_wrap=True)


@click.command()
@click.option(
    '--basedir', '-b', required=True, help='Base directory containing dovecot-uidlist, cur, and tmp directories'
)
@click.option('--maildir', '-m', default=None, help='Path to the Maildir (default: basedir/cur/)')
@click.option(
    '--tmp-dir', '-t', default=None, help='Path to the temporary directory for compression (default: basedir/tmp/)'
)
@click.option(
    '--control-dir',
    '-c',
    default=None,
    help='Path to the control directory containing dovecot-uidlist (default: basedir)',
)
@click.option('--timeout', type=int, default=10, help='Timeout for maildirlock')
@click.option(
    '--compression',
    '-z',
    type=click.Choice(['gzip', 'zstd']),
    default='gzip',
    help='Compression method to use (gzip or zstd)',
)
@click.option('--lock/--no-lock', '-l', is_flag=True, default=True, help='Use maildir locking mechanism')
@click.option('--verbose', is_flag=True, help='Enable verbose output')
@click.version_option(version=__version__)
def main(basedir, maildir, tmp_dir, control_dir, timeout, compression, lock, verbose):
    if maildir is None:
        maildir = basedir + '/cur'
    if tmp_dir is None:
        tmp_dir = basedir + '/tmp'
    if control_dir is None:
        control_dir = basedir

    handler = RichHandler(show_path=False, omit_repeated_times=False, show_level=True)
    formatter = logging.Formatter(fmt='%(message)s', datefmt='[%X]')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    if verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    compressor = MailCompressor(
        maildir=maildir,
        tmp_dir=tmp_dir,
        control_dir=control_dir,
        timeout=timeout,
        compression_method=compression,
        use_lock=lock,
    )
    compressor.run()


if __name__ == '__main__':
    main()



import pytest
import time

from compress_mail.lockmaildir import maildir_lock


def test_maildir_lock(tmp_path):
    lock_path = tmp_path / 'maildir'
    lock_path.mkdir()
    with maildir_lock(lock_path, 10) as lock_file:
        assert lock_file == tmp_path / 'maildir/dovecot-uidlist.lock'
        assert lock_file.exists()
    assert not lock_file.exists()


def test_lock_failure(tmp_path):
    p = tmp_path / 'dovecot-uidlist.lock'
    p.write_text('Test file')
    print(p)
    print(p.exists())
    with pytest.raises(TimeoutError) as excinfo:
        with maildir_lock(tmp_path, 1) as lock_file:
            time.sleep(1)



import os
from pathlib import Path

import pytest
from unittest.mock import patch
from compress_mail.compress_mail import MailCompressor


@patch('os.walk')
def test_find_mails_to_compress(mock_walk):
    mock_walk.return_value = [
        (
            '/path/to/maildir',
            [],
            ['1742423129.M60927P26663.fully,S=30457,W=30918:2,S', '1742423130.M622371P26663.fully,S=12796,W=12984:2,S'],
        )
    ]
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    mails = compressor.find_mails_to_compress()
    assert len(mails) == 2


# @patch('shutil.copy2')
# @patch('subprocess.run')
# def test_compress_mails_gzip(mock_run, mock_copy):
#     compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
#     mails = ['/path/to/maildir/mail1:2,']
#     compressed_files = compressor.compress_mails(mails)
#     assert len(compressed_files) == 1
#     mock_run.assert_called_with(['gzip', '-6', '/tmp/mail1:2,'])


# @patch('os.utime')
# @patch('os.path.getmtime')
# def test_update_mtime(mock_getmtime, mock_utime):
#     mock_getmtime.return_value = 1234567890
#     compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
#     original_files = ['/path/to/maildir/mail1:2,']
#     compressed_files = ['/tmp/mail1:2,.gz']
#     compressor.update_mtime(original_files, compressed_files)
#     mock_utime.assert_called_with('/tmp/mail1:2,.gz', (1234567890, 1234567890))


def test_update_mtime(tmp_path):
    # Create original and compressed files
    original_file = tmp_path / 'original_mail'
    compressed_file = tmp_path / 'compressed_mail.gz'
    original_file.touch()
    compressed_file.touch()

    # Set a specific mtime for the original file
    original_mtime = 1234567890
    os.utime(original_file, (original_mtime, original_mtime))

    # Create a MailCompressor instance
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)

    # Call update_mtime
    compressor.update_mtime([original_file], [compressed_file])

    # Verify that the mtime of the compressed file matches the original file
    assert compressed_file.stat().st_mtime == original_mtime


# @patch('os.rename')
# @patch('os.remove')
# @patch('os.path.exists')
# def test_verify_and_replace_mails(mock_exists, mock_remove, mock_rename):
#     mock_exists.return_value = True
#     compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
#     original_files = ['/path/to/maildir/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa']
#     compressed_files = ['/tmp/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ']
#     compressor.verify_and_replace_mails(original_files, compressed_files)
#     mock_rename.assert_any_call('/tmp/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ', '/path/to/maildir/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa')
#     mock_rename.assert_any_call('/path/to/maildir/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa', '/path/to/maildir/1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ')


def test_verify_and_replace_mails(tmp_path):
    # Create original and compressed files
    original_file = tmp_path / '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa'
    compressed_file = tmp_path / '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa.gz'
    final_file = tmp_path / '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ'
    original_file.touch()
    compressed_file.touch()

    # Create a MailCompressor instance
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)

    # Call verify_and_replace_mails
    compressor.verify_and_replace_mails([original_file], [compressed_file])

    # Verify that the original file has been renamed and the compressed file has been removed
    assert not original_file.exists()
    assert not compressed_file.exists()
    assert final_file.exists()


@pytest.mark.parametrize(
    'path,original,expected',
    [
        (
            '',
            '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa',
            '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ',
        ),
        (
            '.folder.with spaces',
            '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSa',
            '1685154999.M900454P14950.fully,S=13592,W=13909:2,FSaZ',
        ),
    ],
)
def test_run_with_lock(path, original, expected, tmp_path):
    original_file = tmp_path / path / 'cur' / original
    compressed_file = tmp_path / path / 'cur' / expected
    original_file.parent.mkdir(parents=True)
    Path.mkdir(tmp_path / path / 'tmp', parents=True, exist_ok=True)
    original_file.write_text('Test file')
    compressor = MailCompressor(
        maildir=str(tmp_path / path / 'cur'),
        tmp_dir=str(tmp_path / path / 'tmp'),
        control_dir=str(tmp_path / path),
        timeout=10,
        compression_method='gzip',
        use_lock=True,
    )
    compressor.run()
    assert not original_file.exists()
    assert compressed_file.exists()


# @patch('compress_mail.compress_mail.MaildirLock')
# @patch('compress_mail.compress_mail.MailCompressor.check_binaries')
# @patch('compress_mail.compress_mail.MailCompressor.get_directory_size')
# def test_run_without_lock(mock_get_directory_size, mock_check_binaries, mock_lock):
#     mock_get_directory_size.return_value = 1
#     compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
#     compressor.find_mails_to_compress = MagicMock(return_value=['/path/to/maildir/mail1:2,'])
#     compressor.compress_mails = MagicMock(return_value=['/tmp/mail1:2,.gz'])
#     compressor.update_mtime = MagicMock()
#     compressor.verify_and_replace_mails = MagicMock()
#     compressor.run()
#     compressor.verify_and_replace_mails.assert_called_once()


@patch('os.walk')
def test_find_mails_to_compress_returns_empty_list_when_no_mails(mock_walk):
    mock_walk.return_value = [('/path/to/maildir', [], [])]
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    mails = compressor.find_mails_to_compress()
    assert len(mails) == 0


@patch('shutil.copy2')
@patch('subprocess.run')
def test_compress_mails_handles_empty_list(mock_run, mock_copy):
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    compressed_files = compressor.compress_mails([])
    assert len(compressed_files) == 0
    mock_run.assert_not_called()
    mock_copy.assert_not_called()


def test_update_mtime_handles_empty_lists():
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    compressor.update_mtime([], [])
    # No exception should be raised


def test_verify_and_replace_mails_handles_empty_lists():
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    compressor.verify_and_replace_mails([], [])
    # No exception should be raised


def test_with_ending_appends_correctly():
    compressor = MailCompressor('/path/to/maildir', '/tmp', '/control', 10, 'gzip', False)
    path = Path('example.txt')
    new_path = compressor.with_ending(path, 'Z')
    assert new_path.name == 'example.txtZ'



import logging
import os
import signal
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class DotlockSettings:
    timeout: int
    use_excl_lock: bool
    nfs_flush: bool


@contextmanager
def file_dotlock_create(dotlock_settings: DotlockSettings, path: Path) -> Path:
    lock_file = path.with_suffix('.lock')
    start_time = time.time()
    ours = False
    while True:
        try:
            with lock_file.open('x') as f:
                f.write('lock')
            logger.info(f'Created lock file {lock_file}')
            ours = True
            yield lock_file
            break
        except FileExistsError:
            logger.info(f'Lock file {lock_file} already exists')
            if time.time() - start_time >= dotlock_settings.timeout:
                raise TimeoutError(f"Failed to acquire lock within {dotlock_settings.timeout} seconds")
            time.sleep(1)
        finally:
            if lock_file.exists() and ours:
                logger.info(f'Removing lock file {lock_file}')
                lock_file.unlink()


def maildir_lock(path: Path, timeout: int) -> Path:
    dotlock_settings = DotlockSettings(
        timeout=timeout,
        use_excl_lock=os.getenv('DOTLOCK_USE_EXCL') is not None,
        nfs_flush=os.getenv('MAIL_NFS_STORAGE') is not None,
    )
    lock_path = path / 'dovecot-uidlist'
    return file_dotlock_create(dotlock_settings, lock_path)


def sig_die(signum, frame):
    print('Signal received, exiting...')
    sys.exit(0)


def main():
    if len(sys.argv) != 3:
        print('Usage: maildirlock <path> <timeout>\n - SIGTERM will release the lock.')
        sys.exit(1)

    path = Path(sys.argv[1])
    try:
        timeout = int(sys.argv[2])
    except ValueError:
        print(f'Invalid timeout value: {sys.argv[2]}')
        sys.exit(1)

    signal.signal(signal.SIGTERM, sig_die)
    signal.signal(signal.SIGINT, sig_die)

    lock_acquired = False
    try:
        with maildir_lock(path, timeout) as lock_file:
            lock_acquired = True
            pid = os.getpid()
            print(pid)
            while True:
                time.sleep(1)
    except Exception as e:
        if not lock_acquired:
            print(f'Failed to acquire lock: {e}')
        sys.exit(1)


if __name__ == '__main__':
    main()
